
unit S_FolderEx;

interface

uses
  Windows, Messages, Classes, Controls, ShlObj, ActiveX;

type
  TFolderEx = class(TComponent)
  private
    FOwner: TObject;
    FTitle: string;
    FComment: string;
    FDefaultFolder: string;
    function GetFolderFromCommonDlg(AComment: string): string;
    procedure SetWindow_Pos_Size;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute: string;
    property Title: string read FTitle write FTitle;
    property Comment: string read FComment write FComment;
    property DefaultFolder: string read FDefaultFolder write FDefaultFolder;
  end;

var
  FolderEx: TFolderEx = nil;

implementation

const
  ID_EDIT   = $3744;  // Edit
  ID_NEWBTN = $3746;  // Button

var
  BrowseInfo: TBrowseInfo;
  hDlg: THandle = 0;
  hEdit: THandle = 0;
  hNewBtn: THandle = 0;
  PathName: array[0..MAX_PATH] of Char;
  PathNameA: array[0..MAX_PATH] of Char;
  OrgProc: FARPROC = nil;
  DidShow: Boolean = False;
  TextBoxChanged: Boolean = False;

function SubProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): Integer; stdcall;
begin
	case uMsg of
    WM_SHOWWINDOW:
      if not DidShow and (FolderEx <> nil) then
      begin
        DidShow := True;
        FolderEx.SetWindow_Pos_Size;
      end;
  end;
  Result := CallWindowProc(OrgProc, hWnd, uMsg, wParam, lParam);
end;

function BrowseCallbackProc(hWnd: HWND; uMsg: UINT; lParam: LPARAM; lpData: LPARAM): Integer; stdcall;
var
  R: TRect;
  Pt: TPoint;
begin
  case uMsg of
    BFFM_INITIALIZED:
      begin
        hDlg := hWnd;
        hEdit := GetDlgItem(hDlg,ID_EDIT);
        hNewBtn := GetDlgItem(hDlg,ID_NEWBTN);
        if hNewBtn > 0 then
        begin
          SetWindowText(hNewBtn, 'VK쐬(&N)');
          GetWindowRect(hNewBtn, R);
          Pt := Point(R.Left, R.Top);
          Windows.ScreenToClient(hDlg, Pt);
          MoveWindow(hNewBtn, Pt.X, Pt.Y, 100, R.Bottom - R.Top, False);
        end;
        SendMessage(hWnd, BFFM_SETSELECTION, WPARAM(TRUE), lpData);
        OrgProc := Pointer(GetWindowLong(hDlg, GWL_WNDPROC));
        SetWindowLong(hDlg, GWL_WNDPROC, NativeInt(@SubProc));
      end;
    BFFM_SELCHANGED:
      if BrowseInfo.ulFlags and BIF_EDITBOX = BIF_EDITBOX then
      begin
        TextBoxChanged := False;
        SHGetPathFromIDList(PItemIDList(lParam), PathName);
        if (hEdit > 0) and (PathName <> '') then
        SendMessage(hEdit, WM_SETTEXT, 0, NativeInt(@PathName));
      end;
    BFFM_VALIDATEFAILED:
      if BrowseInfo.ulFlags and BIF_EDITBOX = BIF_EDITBOX then
      begin
        TextBoxChanged := True;
        Windows.GetWindowText(hEdit, PathNameA, MAX_PATH);
      end;
  end;
  Result := 0;
end;

constructor TFolderEx.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FOwner := AOwner;
  FTitle := 'tH_[̎Q';
  FComment := '';
  FDefaultFolder := 'C:\';
  DidShow := False;
  TextBoxChanged := False;
end;

destructor TFolderEx.Destroy;
begin
  SetWindowLong(hDlg, GWL_WNDPROC, NativeInt(OrgProc));
  inherited Destroy;
end;

procedure TFolderEx.SetWindow_Pos_Size;
var
  R0,R: TRect;
  W0,H0,W,H,L,T: Integer;
begin
  if hDlg > 0 then
  begin
    Sleep(200);
    SystemParametersInfo(SPI_GETWORKAREA, 0, R0, 0);
    GetWindowRect(hDlg, R);
    SetWindowText(hDlg, FTitle);
    W0 := R0.Right - R0.Left;
    H0 := R0.Bottom - R0.Top;
    W := R.Right - R.Left;
    H := R.Bottom - R.Top;
    if W > W0 then
      W := W0;
    if H > H0 then
      H := H0;
    L := (W0 - W) div 2;
    T := (H0 - H) div 2;
    if L < R0.Left then
      L := R0.Left;
    if T < R0.Top then
      T := R0.Top;
    SetWindowPos(hDlg, HWND_TOP, L, T, W, H, SWP_SHOWWINDOW);
  end;
end;

function TFolderEx.Execute: string;
begin
  Result := '';
  hDlg := 0;
  Result := GetFolderFromCommonDlg(FComment);
end;

function TFolderEx.GetFolderFromCommonDlg(AComment: string): string;
var
  FolderPidl: PItemIDList;
  AMalloc: IMalloc;
begin
  Result := '';
  ShGetMalloc(AMalloc);
  FolderPidl := nil;
  FillChar(BrowseInfo, SizeOf(BrowseInfo), #0);
  BrowseInfo.hwndOwner := TWinControl(FOwner).Handle;
  BrowseInfo.lpfn      := @BrowseCallbackProc;
  BrowseInfo.lpszTitle := PChar(AComment);
  BrowseInfo.pidlRoot  := nil;
  BrowseInfo.lParam    := LPARAM(PChar(FDefaultFolder));
  BrowseInfo.ulFlags   := BIF_RETURNONLYFSDIRS or BIF_EDITBOX or BIF_VALIDATE or
                          BIF_NEWDIALOGSTYLE or BIF_BROWSEFORCOMPUTER;
  FolderPidl := SHBrowseForFolder(BrowseInfo);
  if FolderPidl <> nil then
  try
    if TextBoxChanged then
      Result := PathNameA
    else
    begin
      SHGetPathFromIDList(FolderPidl, PathName);
      Result := PathName;
    end;
  finally
    AMalloc.Free(FolderPidl);
  end;
end;

end.

